--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose: Provide streaming interface to DDR memory
-- Description:
--   Write or read a block of data to or from DDR memory. The data width is set
--   by the DDR controller data width given by RESIZE_DDR_CTLR_DATA() and eg.
--   256 bits for DDR3 with 64 bit DQ data. The block of data is located from
--   dvr_start_addr to dvr_end_addr.
--   The io_ddr_driver takes care that the access is done in a number of bursts.
--   The write burst size depends on the maximum burst size, the remaining
--   block size and on the number of words available in the (external) FIFO as
--   indicated by wr_fifo_usedw. The read burst size depends on the maximum
--   burst size and the remaining block size.
-- Remarks:
-- . 

LIBRARY IEEE, tech_ddr_lib, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;

ENTITY io_ddr_driver IS 
  GENERIC (
    g_tech_ddr         : t_c_tech_ddr
  );
  PORT ( 
    rst                : IN  STD_LOGIC;
    clk                : IN  STD_LOGIC;

    dvr_en             : IN  STD_LOGIC := '1';
    dvr_wr_not_rd      : IN  STD_LOGIC;       
    dvr_start_address  : IN  STD_LOGIC_VECTOR;
    dvr_end_address    : IN  STD_LOGIC_VECTOR;
    dvr_done           : OUT STD_LOGIC;         -- Requested wr or rd sequence is done.
   
    wr_fifo_usedw      : IN  STD_LOGIC_VECTOR;
    wr_snk_in          : IN  t_dp_sosi;
    wr_snk_out         : OUT t_dp_siso;
    
    rd_src_out         : OUT t_dp_sosi;
    rd_src_in          : IN  t_dp_siso;
    
    ctlr_init_done     : IN  STD_LOGIC;
    ctlr_miso          : IN  t_tech_ddr_miso;
    ctlr_mosi          : OUT t_tech_ddr_mosi
   );
END io_ddr_driver;


ARCHITECTURE str OF io_ddr_driver IS

  CONSTANT c_ctlr_address_w     : NATURAL := func_tech_ddr_ctlr_address_w(g_tech_ddr);
 
  CONSTANT c_margin             : NATURAL := 2; -- wr_burst_size is updated one cycle after reading actual nof available words.
                                                -- Subtract two (wr_fifo_usedw and wr_burst_size are both registered) so we cannot 
                                                -- post a request for a too large burst size, which could cause the wr_burst state 
                                                -- to be two valid words short.

  SIGNAL req_burst_cycles       : STD_LOGIC_VECTOR(g_tech_ddr.maxburstsize_w-1 DOWNTO 0);
  SIGNAL nxt_req_burst_cycles   : STD_LOGIC_VECTOR(g_tech_ddr.maxburstsize_w-1 DOWNTO 0);

  TYPE t_state_enum IS (s_init, s_idle, s_wait1, s_wait2, s_wait3, s_rd_request, s_wr_request, s_wr_burst);

  SIGNAL state                  : t_state_enum;
  SIGNAL nxt_state              : t_state_enum; 
  SIGNAL prev_state             : t_state_enum; 

  SIGNAL wr_burst_size          : NATURAL;
  SIGNAL rd_burst_size          : NATURAL;

  SIGNAL nxt_wr_burst_size      : NATURAL;
  SIGNAL nxt_rd_burst_size      : NATURAL;

  SIGNAL i_dvr_done             : STD_LOGIC;
  SIGNAL nxt_dvr_done           : STD_LOGIC;

  SIGNAL cur_address            : STD_LOGIC_VECTOR(c_ctlr_address_w-1 DOWNTO 0);
  SIGNAL nxt_cur_address        : STD_LOGIC_VECTOR(c_ctlr_address_w-1 DOWNTO 0);   
  SIGNAL diff_address           : STD_LOGIC_VECTOR(c_ctlr_address_w-1 DOWNTO 0);

  SIGNAL addresses_rem          : STD_LOGIC_VECTOR(c_ctlr_address_w-1 DOWNTO 0); -- nof words (on the user side interface) to rd/wr until end addr is reached
  SIGNAL reg_addresses_rem      : STD_LOGIC_VECTOR(c_ctlr_address_w-1 DOWNTO 0); -- nof words (on the user side interface) to rd/wr until end addr is reached

  SIGNAL reg_wr_fifo_usedw      : STD_LOGIC_VECTOR(wr_fifo_usedw'RANGE);  -- read side depth of the write FIFO

BEGIN

  dvr_done <= i_dvr_done; 

  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      state                <= s_init;
      req_burst_cycles     <= (OTHERS => '0');
      i_dvr_done           <= '0';
      cur_address          <= (OTHERS=>'0');
      wr_burst_size        <= 0;
      rd_burst_size        <= 0;
      reg_addresses_rem    <= (OTHERS=>'0');
      reg_wr_fifo_usedw    <= (OTHERS=>'0');
      prev_state           <= s_idle;
    ELSIF rising_edge(clk) THEN
      state                <= nxt_state;
      req_burst_cycles     <= nxt_req_burst_cycles;
      i_dvr_done           <= nxt_dvr_done;
      cur_address          <= nxt_cur_address;
      wr_burst_size        <= nxt_wr_burst_size;
      rd_burst_size        <= nxt_rd_burst_size;
      reg_addresses_rem    <= addresses_rem;
      reg_wr_fifo_usedw    <= wr_fifo_usedw;
      prev_state           <= state;
    END IF;
  END PROCESS;

  -- End address - current address
  diff_address  <= SUB_UVEC(dvr_end_address, cur_address);

  -- Add 1 to diff_address to also write the last address
  addresses_rem <= INCR_UVEC(diff_address, 1); 
  
  p_wr_burst_size : PROCESS (reg_addresses_rem, reg_wr_fifo_usedw)
    VARIABLE v_burst_size : POSITIVE;
  BEGIN 
    -- Write burst size at least 1 and if more then set to the smallest of g_tech_ddr.maxburstsize, addresses_rem and wr_fifo_usedw
    IF UNSIGNED(reg_wr_fifo_usedw)>c_margin AND UNSIGNED(reg_addresses_rem)>c_margin THEN 
      v_burst_size := g_tech_ddr.maxburstsize+c_margin;
      IF v_burst_size > UNSIGNED(reg_addresses_rem) THEN v_burst_size := TO_UINT(reg_addresses_rem); END IF;
      IF v_burst_size > UNSIGNED(reg_wr_fifo_usedw) THEN v_burst_size := TO_UINT(reg_wr_fifo_usedw); END IF;
      v_burst_size := v_burst_size - c_margin;
    ELSE
      v_burst_size := 1;
    END IF;
    nxt_wr_burst_size <= v_burst_size;
  END PROCESS;

  p_rd_burst_size : PROCESS (reg_addresses_rem)
    VARIABLE v_burst_size : POSITIVE;
  BEGIN 
    -- Read burst size is at least 1 and if more then set to the smallest of g_tech_ddr.maxburstsize and addresses_rem
    IF UNSIGNED(reg_addresses_rem)>0 THEN
      v_burst_size := g_tech_ddr.maxburstsize;
      IF v_burst_size > UNSIGNED(reg_addresses_rem) THEN v_burst_size := TO_UINT(reg_addresses_rem); END IF;
    ELSE
      v_burst_size := 1;
    END IF;
    nxt_rd_burst_size <= v_burst_size;
  END PROCESS;
  
  rd_src_out.valid <= ctlr_miso.rdval;
  rd_src_out.data <= RESIZE_DP_DATA(ctlr_miso.rddata);
  
  p_state : PROCESS(prev_state, state, ctlr_init_done,
                    dvr_en, dvr_wr_not_rd, i_dvr_done, ctlr_miso, req_burst_cycles, wr_snk_in, rd_src_in, 
                    wr_fifo_usedw, wr_burst_size, rd_burst_size, reg_addresses_rem, dvr_start_address, cur_address)
  BEGIN  
    nxt_state              <= state;
    
    ctlr_mosi.address      <= RESIZE_DDR_CTLR_ADDRESS(cur_address);
    ctlr_mosi.wrdata       <= RESIZE_DDR_CTLR_DATA(wr_snk_in.data);
    ctlr_mosi.wr           <= '0';
    ctlr_mosi.rd           <= '0';
    ctlr_mosi.burstbegin   <= '0'; 
    ctlr_mosi.burstsize    <= (OTHERS => '0');
    
    wr_snk_out.ready       <= '0';
    nxt_req_burst_cycles   <= req_burst_cycles;    
    nxt_dvr_done           <= i_dvr_done;
    nxt_cur_address        <= cur_address;

    CASE state IS
     
      WHEN s_wr_burst => -- Performs the burst portion (word 2+)        
        ctlr_mosi.wr <= '1';        
        IF ctlr_miso.waitrequest_n = '1' THEN
          nxt_req_burst_cycles  <= INCR_UVEC(req_burst_cycles, -1);
          wr_snk_out.ready  <= '1'; -- wr side uses latency of 0, so wr_snk_out.ready<='1' acknowledges a successful write request.
          IF UNSIGNED(req_burst_cycles) = 1 THEN -- Then we're in the last cycle of this burst sequence
            nxt_state <= s_wr_request; -- We can only initiate a new burst through the wr_request state
          END IF;
        END IF;            

      WHEN s_wr_request => -- Performs 1 write or read and goes into s_wr_burst when requested write words >1         
        nxt_state <= s_wait3; 
        IF UNSIGNED(reg_addresses_rem) = 0 THEN -- end address reached
          nxt_dvr_done  <= '1';              
          nxt_state     <= s_idle;          
        ELSIF ctlr_miso.waitrequest_n = '1' THEN 
          IF wr_snk_in.valid = '1' THEN
            -- Always perform 1st write here             
            wr_snk_out.ready     <= '1';
            ctlr_mosi.wr         <= '1';
            ctlr_mosi.burstbegin <= '1';                                   -- assert burstbegin,
            ctlr_mosi.burstsize  <= TO_DDR_CTLR_BURSTSIZE(wr_burst_size);  -- burstsize >= 1
            IF wr_burst_size > 1 THEN
              -- Perform any remaining writes in a burst
              nxt_state            <= s_wr_burst;
              nxt_req_burst_cycles <= TO_DDR_CTLR_BURSTSIZE(wr_burst_size-1); -- first burst wr cycle is done here, the rest are done in s_wr_burst
            END IF; -- ELSE: there is only 1 word, so no need for remaining burst   
            nxt_cur_address   <= INCR_UVEC(cur_address, wr_burst_size);
          END IF;
        END IF;        

      WHEN s_rd_request =>  -- Posts a read request for a burst (1...g_tech_ddr.maxburstsize)      
        nxt_state <= s_wait3;   
        IF UNSIGNED(reg_addresses_rem) = 0 THEN -- end address reached
          nxt_dvr_done  <= '1';              
          nxt_state     <= s_idle;
        ELSE 
          IF rd_src_in.ready = '1' THEN -- Fifo uses its internal almost_full signal to toggle its snk_out.rdy     
            IF ctlr_miso.waitrequest_n = '1' THEN    
              ctlr_mosi.rd         <= '1';                   
              ctlr_mosi.burstbegin <= '1';                                   -- assert burstbegin,
              ctlr_mosi.burstsize  <= TO_DDR_CTLR_BURSTSIZE(rd_burst_size);  -- burstsize >= 1
              nxt_cur_address   <= INCR_UVEC(cur_address, rd_burst_size);  
            END IF;           
          END IF;
        END IF; 

      -- This wait state is inserted between two requests when necessary, e.g. when FSM enters wr_request
      -- from the state wr_request, an extra cycle is needed for reg_addresses_rem to be valid.
      WHEN s_wait3 =>
        IF prev_state = s_wr_request THEN nxt_state <= s_wr_request; END IF;
        IF prev_state = s_rd_request THEN nxt_state <= s_rd_request; END IF;
   
     -- In this cycle reg_addresses_rem is valid. This cycle is added so wr_burst_size and rd_burst_size
     -- (derived from reg_addresses_rem) are valid the next cycle.
     WHEN s_wait2 =>
       IF dvr_wr_not_rd = '1' THEN
         nxt_state <= s_wr_request; 
       ELSE
         nxt_state <= s_rd_request;             
       END IF;
      
      -- Wait a cycle so reg_addresses_rem is valid the next cyle.
      WHEN s_wait1 =>
        nxt_state <= s_wait2;

      WHEN s_idle =>
        nxt_dvr_done <= '1';    -- assert dvr_done after s_init or keep it asserted after a finished access
        IF dvr_en = '1' THEN  
          nxt_cur_address <= dvr_start_address; 
          nxt_dvr_done    <= '0';
          nxt_state       <= s_wait1;
        END IF;
     
      WHEN OTHERS => -- s_init
        nxt_dvr_done <= '0';
        IF ctlr_init_done = '1' THEN
          nxt_state <= s_idle;  -- and assert dvr_done when in s_idle to indicate ctlr_init_done
        END IF;
      
    END CASE;
  END PROCESS;

END str;

